#######################################################################
#                       Character Encoding                            #
#######################################################################

def ez_encode_str(to_encode: str) -> list:
    """ Encodes a `str` per character and puts them into a list.

    :type to_encode: str
    :param to_encode: The string that has to be encoded.

    :rtype: list
    :return: A list of encoded chars. Encodes in UTF-8
    """
    to_return = []
    for char in list(to_encode):
        if type(char) != bytes:
            try:
                to_return.append(char.encode("utf-8"))
            except AttributeError:
                raise Exception("`to_encode` must be of type `str`")
        else:
            # This is a problem as they could be encoded differently.
            to_return.append(char)
    return to_return


#######################################################################
#                            Vi tools                                 #
#######################################################################


# Writing

def write_chars(to_write) -> list:
    """To type ``to_write`` to the file.

    ``write_chars`` will type the passed string after the cursor position. 
    From Vi’s command mode, it types ``a`` to insert after and then types the 
    string.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_chars: "snek"

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_chars("snek")

    :type to_write: str
    :param to_write: The characters to write.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    to_write = "a" + to_write + chr(27)
    to_write = ez_encode_str(to_write)

    return to_write


def write_line(to_write):
    """To type `to_write` and create a new line.

    Starts typing after the current cursor position by pressing
    ``a`` from the command mode. ``to_write`` is then typed and
    a new line is created. 

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_line: "Python is fun."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_line("Python is fun.")

    :type to_write: str
    :param to_write: The characters to write.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    to_write = "a" + to_write + "\n" + chr(27)
    to_write = ez_encode_str(to_write)

    return to_write


def new_line(amount=1):
    """Creates an ``amount`` of new lines.

    ``new_line`` inserts a certain number of new lines to the file. 
    From Vi’s command mode, ``ezvi`` first presses ``o``. This ensures 
    that the current line won’t be split even if the cursor is not 
    at the end of the line.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - new_line: 3


    `Using the API:`

    .. code-block:: python

      ezvi.tools.new_line(3)

    :type amount: int
    :param amount: The number of new lines to create.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """
    if type(amount) != int:
        try:
            amount = int(amount)
        except TypeError:
            amount = 1

    to_write = "o" + "\n" * (amount-1) + chr(27)
    to_write = ez_encode_str(to_write)

    return to_write


def new_line_over():
    """Creates a new line over the cursor.

    The cursor is also moved to the beginning of the new line. It is
    not possible to create more than one new line over the cursor
    at a time for now.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - new_line_over:


    `Using the API:`

    .. code-block:: python

      ezvi.tools.new_line_over()

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    to_write = "O" + chr(27)
    to_write = ez_encode_str(to_write)

    return to_write


def write_after_word(to_write):
    """To write ``to_write`` after the current word.

    This function uses ``e`` from the command mode to go to the end
    of the word. ``to_write`` is then written after the end of the
    word using the ``a`` command. **This function does not add a
    space to the beginning of ``to_write``.**

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_after_word: " General Kenobi."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_after_word(" General Kenobi.")

    :type to_write: str
    :param to_write: What to write after the word.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    prepend = "e" + "a"
    append = chr(27)
    to_write = prepend + to_write + append
    to_write = ez_encode_str(to_write)

    return to_write


def write_after_line(to_write):
    """To write ``to_write`` after the current line.

    This function uses ``$`` from the command mode to go to the 
    end of the line. ``to_write`` is then written after the cursor
    position using the ``a`` command. **This function does not add a
    space to the beginning of ``to_write``.**

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_after_line: " General Kenobi."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_after_line(" General Kenobi.")

    :type to_write: str
    :param to_write: What to write after the line.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    prepend = "$" + "a"
    append = chr(27)
    to_write = prepend + to_write + append
    to_write = ez_encode_str(to_write)

    return to_write


def write_after_char(to_write):
    """To write ``to_write`` after the cursor's position (current char).

    ``to_write`` is written after the cursor
    position using the ``a`` command. 

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_after_char: "Greetings!"

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_after_char("Greetings!")

    :type to_write: str
    :param to_write: What to write after the cursor.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    prepend = "a"
    append = chr(27)
    to_write = prepend + to_write + append
    to_write = ez_encode_str(to_write)

    return to_write


def write_before_word(to_write):
    """To write ``to_write`` before the current word.

    ``write_before_word`` uses ``b`` from the command mode to move
    the cursor to the beginning of the current word. ``to_write`` is
    then written before the cursor's position using the ``i`` 
    command.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_before_word: "Hello there."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_before_word("Hello there.")

    :type to_write: str
    :param to_write: What to write before the current word.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    prepend = "b" + "i"  # TODO: Replace "b" by something that works.
    append = chr(27)
    to_write = prepend + to_write + append
    to_write = ez_encode_str(to_write)

    return to_write


def write_before_line(to_write):
    """To write ``to_write`` at the beginning of the line.

    ``write_before_line`` uses ``0`` from the command mode to move
    the cursor to the beginning of the current line. ``to_write`` is
    then written before the cursor's position using the ``i`` 
    command.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_before_line: "Hello there."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_before_line("Hello there.")

    :type to_write: str
    :param to_write: What to write before the current line.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    prepend = "0" + "i"
    append = chr(27)
    to_write = prepend + to_write + append
    to_write = ez_encode_str(to_write)

    return to_write


def write_before_char(to_write):
    """To write ``to_write`` at the beginning of the line.

    ``write_before_line`` uses ``0`` from the command mode to move
    the cursor to the beginning of the current line. ``to_write`` is
    then written before the cursor's position using the ``i`` 
    command.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_before_line: "Hello there."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_before_line("Hello there.")

    :type to_write: str
    :param to_write: What to write before the current line.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    prepend = "i"
    append = chr(27)
    to_write = prepend + to_write + append
    to_write = ez_encode_str(to_write)

    return to_write


# Movement

def goto_line(line_num):
    """To go to a certain line.

    This function uses the ``G`` command to move the cursor to
    the beginning of a certain line.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - goto_line: 5

    `Using the API:`

    .. code-block:: python

      ezvi.tools.goto_line(5)

    :type line_num: int
    :param line_num: The number of the line to move the cursor to.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    to_write = str(line_num) + "G"
    to_write = ez_encode_str(to_write)

    return to_write


def goto_column(column_num):
    """To go to a certain column.

    This function uses the ``l`` command to move the cursor to
    a certain column on the **current line**.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - goto_column: 5

    `Using the API:`

    .. code-block:: python

      ezvi.tools.goto_column(5)

    :type column_num: int
    :param column_num: The number of the column to move the cursor to.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    # This would be much cleaner if I could get the cursor's position.
    to_write = "0" + str(column_num - 1) + "l"
    to_write = ez_encode_str(to_write)

    return to_write


# Replace functions

def replace(start, end, new):
    """Replaces text on the current line.

    This function replaces from the column number ``start`` to the
    column number ``end`` with the ``new`` text. ``replace`` moves
    the cursor to the starting position and then uses the change 
    command (``c``) to replace the text.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - replace: 0, 4, "Snek"

    `Using the API:`

    .. code-block:: python

      ezvi.tools.replace(0, 4, "Snek")

    :type start: int
    :param start: The number of the column to start replacing (inclusively).

    :type end: int
    :param end: The number of the column to stop replacing (exlusively).

    :type new: str
    :param new: The new text to type.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    movement = goto_column(start)
    replace = "c" + str(end - start)
    to_write = movement + replace + new + chr(27)
    to_write = ez_encode_str(to_write)

    return to_write


def find_replace(old, new):
    """To find `old` on the current line and replaces it with `new`."""

    pass


def replace_line(new):
    """Replaces text on the current line.

    ``replace`` moves the cursor to the beginning of the line 
    using ``0`` from the command mode and then uses command 
    (``c$``) to replace the whole line.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - replace_line: "Hello there."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.replace_line("General Kenobi.")

    :type new: str
    :param new: The new text to replace the current line.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    movement = "0"
    replace = "c" + "$"
    to_write = movement + replace + new + chr(27)
    to_write = ez_encode_str(to_write)

    return to_write


# Vi commands

def write_file(filename):
    """Write the current buffer.

    From the command mode, ``write_file`` uses Vi's ``:w`` command to
    write the current buffer.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_file: "message.txt"

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_file("message.txt")

    :type filename: str
    :param filename: The path towards where to save the current buffer.

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    to_write = ":w " + filename + "\n"
    to_write = ez_encode_str(to_write)

    return to_write


def quit_editor():
    """To quit the editor.

    From the command mode, ``quit_editor`` uses Vi's ``:q`` command to
    quit the editor. Since Vi makes sure that the current buffer
    is saved before quitting, ``force_quit_editor`` should be used instead
    if the buffer must not be saved.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - quit_editor:

    `Using the API:`

    .. code-block:: python

      ezvi.tools.quit_editor()

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """

    to_write = ":q" + "\n"
    to_write = ez_encode_str(to_write)

    return to_write


def force_quit_editor():
    """To force quit the editor.

    From the command mode, ``force_quit_editor`` uses Vi's ``:q!`` command to
    force quit the editor. All unsaved progress will be lost.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - force_quit_editor:

    `Using the API:`

    .. code-block:: python

      ezvi.tools.force_quit_editor()

    :rtype: list
    :return: A list of encoded characters that can be directly interpreted by ``Vi``.
    """
    """To force quit the editor.

    :rtype: list
    """

    to_write = ":q!" + "\n"
    to_write = ez_encode_str(to_write)

    return to_write


all_tools = {
    write_chars.__name__: write_chars,
    write_line.__name__: write_line,
    new_line.__name__: new_line,
    new_line_over.__name__: new_line_over,
    write_after_word.__name__: write_after_word,
    write_after_line.__name__: write_after_line,
    write_after_char.__name__: write_after_char,
    write_before_word.__name__: write_before_word,
    write_before_line.__name__: write_before_line,
    write_before_char.__name__: write_before_char,
    goto_line.__name__: goto_line,
    goto_column.__name__: goto_column,
    replace.__name__: replace,
    replace_line.__name__: replace_line,
    write_file.__name__: write_file,
    quit_editor.__name__: quit_editor,
    force_quit_editor.__name__: force_quit_editor,
}



import os
import pty
import time
from typing import Dict, Any, Callable
from pathlib import Path
from ezvi import tools
import yaml
import tty
from select import select

STDIN_FILENO = 0
STDOUT_FILENO = 1
CHILD = 0


#######################################################################
#                         Read function                               #
#######################################################################

def ez_read(fd) -> bytes:
    """ Standard read function.

    :type fd: int
    :param fd: File descriptor.

    :rtype: bytes
    :return: Up to 1024 bytes that have been read from `fd`.
    """
    return os.read(fd, 1024)


#######################################################################
#                Spawning an writing to process                       #
#######################################################################

def ez_spawn(argv, instructions, master_read=ez_read, stdin_read=ez_read):
    """ Spawns a process
    To spawn the process. Heavily inspired from Python's `pty`
    module. `ez_spawn()` should only be used once per file that
    needs to be edited. This function can write a whole text file
    following the instructions dictionary and executing those
    instructions using `ez_copy()`.

    :type argv: tuple
    :param argv: First element is the program to run. Other elements
    are the arguments passed to the program.
    :type instructions: list
    :param instructions: Contains all the instructions that will be
    passed to Vi. Its contents should have been created using the
    "Vi tools".
    :type master_read: function
    :param master_read: The function that will be used to read
    info from the master's file descriptor.
    :type stdin_read: function
    :param stdin_read: The function that will be used to read
    info from the master's file descriptor.

    :rtype: int
    :return: Exit status
    """
    all_written = []  # Useful for debugging.
    if type(argv) == str:
        argv = (argv,)
    pid, master_fd = pty.fork()
    if pid == CHILD:
        # Fork worked, run the program
        os.execlp(argv[0], *argv)
    try:
        mode = tty.tcgetattr(STDIN_FILENO)
        # The next line is required to make sure that you can't type over vi
        # either.
        tty.setraw(STDIN_FILENO)  # disable line buffering
        # interrupt signals are no longer interpreted
        restore = 1
    except tty.error:
        # Did not work, no need to restore.
        restore = 0
    # This is where the fun begins.
    # Encoding the instructions.
    try:
        for item in instructions:
            all_written.append(ez_write(master_fd,
                                        item,
                                        master_read,
                                        stdin_read))
    except OSError:
        if restore:
            # Discard queued data and change mode to original.
            tty.tcsetattr(STDIN_FILENO, tty.TCSAFLUSH, mode)
    if restore:
        tty.tcsetattr(STDIN_FILENO, tty.TCSAFLUSH, mode)
    os.close(master_fd)
    # wait for completion and return exit status
    return os.waitpid(pid, 0)[1]


def ez_write(master_fd, to_write, master_read=ez_read, stdin_read=ez_read):
    """Writes every char in `to_write` to `master_fd`.

    :type master_fd: int
    :param master_fd: Master's file descriptor.
    :type to_write: list
    :param to_write: List of encoded chars that will be
    written to `master_fd`.
    :type master_read: function
    :param master_read: The function that will be used to read
    info from the master's file descriptor.
    :type stdin_read: function
    :param stdin_read: The function that will be used to read
    from STDIN (if the user wants to write to the program).

    :rtype: list
    :return: A list of all chars that have been written.
    """
    written = []
    fds = [master_fd, STDIN_FILENO]
    while True:
        rfds, wfds, xfds = select([fds[0]], [fds[1]], [])
        if master_fd in rfds:
            # This is required to see the program running.
            data = master_read(master_fd)
            if not data:
                fds.remove(master_fd)
            else:
                # Printing the program
                os.write(STDOUT_FILENO, data)
        # This should always be true.
        if STDIN_FILENO in wfds:
            try:
                data = to_write.pop(0)  # The item should already be encoded.
            except IndexError:
                data = None
            if not data:
                break
            else:
                # This should be randomized to simulate typing.
                os.write(master_fd, data)
                written.append(data)
                time.sleep(.1)
    return written


#######################################################################
#                            YAML parsing                             #
#######################################################################

def yaml_parser(stream) -> list:
    """Loads a YAML file. 

    :type stream: textIO
    :param stream: A stream of text to be parsed.

    :rtype: list
    :return: The parsed yaml file.
    """

    available = [key for key, value in tools.all_tools.items()]

    parsed = yaml.safe_load(stream)
    for instruction in parsed:
        for key, value in instruction.items():
            # If the function is available, run it with "value"
            # as an argument. This translates the text (values)
            # according to the instructions (values).
            if key in available:
                # Removing None types from dict.
                if instruction[key] is not None:
                    instruction[key] = (tools.all_tools[key](value))
                else:
                    instruction[key] = (tools.all_tools[key]())
            else:
                raise NotImplementedError(key + " does not exist.")

    return parsed


#######################################################################
#                       File comparison                               #
#######################################################################

# To convert an existing text file to some ezvi commands.

def path_check(name):
    """To check if a path to save is valid.

    :param name: The name of the file.
    :type name: str

    :raises Exception: If the path is not valid.

    :return: True if the path is valid.
    :rtype: bool
    """
    if Path(name).is_file():
        raise Exception("{} already exists.".format(name))
    else:
        return True

def file_parser(stream, name=""):
    """To parse a pre-typed text file.

    :param stream: The stream of text to parse.
    :type stream: textIO

    :param save: Where the file should be saved.
    :type save: str

    :return: The parsed text.
    :rtype: list
    """

    to_return = []

    file = stream.readlines()
    for line in file:
        to_return.append(tools.write_chars(line))

    if name == "":
        to_return.append(tools.force_quit_editor())
    else:
        if path_check(name):
            to_return.append(tools.write_file(name))
            to_return.append(tools.quit_editor())

    return to_return

def new_conf(stream, savepath):
    """To create a new configuration file.

    :param stream: The stream of text to create a config from.
    :type stream: textIO
    :param savepath: The path to save the new config file.
    :type savepath: str
    :return: None
    :rtype: NoneType
    """
    
    to_write = []

    if path_check(savepath):
        file = open(savepath, 'w')

    for line in stream:
        line = line.strip()
        if not line:
            to_append = "- new_line: "
        else:
            to_append = "- write_line: " + line + "\n"
        to_write.append(to_append)

    file.writelines(to_write)

    return None

#######################################################################
#                      Searching/editing tools                        #
#######################################################################
